//**********************************************************************
// CNTService
// ==========
//**********************************************************************

#ifndef STRICT
#define STRICT
#endif

#include <windows.h>
#include <stdio.h>
#include "CNTService.h"


//**********************************************************************
// Function prototypes
//**********************************************************************

const char* GetLastErrorMessage(void);


//**********************************************************************
// Static variables
//**********************************************************************

CNTService* CNTService::m_pThis = NULL;

#define PATHKEY   "SYSTEM\\CurrentControlSet\\Services\\RhsAdminSvc\\Parameters"
#define PATHVALUE "Path"


//**********************************************************************
// Constructor
// ===========
//**********************************************************************

CNTService::CNTService(const char* szServiceName)
{

// Copy the address of the current object so we can access it from
// the static member callback functions.
// WARNING: This limits the application to only one CNTService object.

  m_pThis = this;

// Set the default service name and version

  lstrcpyn(m_szServiceName, szServiceName, sizeof(m_szServiceName)-1);
  m_iMajorVersion = 1;
  m_iMinorVersion = 0;
  m_hEventSource = NULL;

// set up the initial service status

  m_hServiceStatus = NULL;
  m_Status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  m_Status.dwCurrentState = SERVICE_STOPPED;
  m_Status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
  m_Status.dwWin32ExitCode = 0;
  m_Status.dwServiceSpecificExitCode = 0;
  m_Status.dwCheckPoint = 0;
  m_Status.dwWaitHint = 0;
  m_bIsRunning = FALSE;
}


CNTService::~CNTService()
{
  if (m_hEventSource)
    ::DeregisterEventSource(m_hEventSource);
}


//**********************************************************************
// ParseStandardArgs
// =================
// Parse the command line.
// Returns TRUE if it found an arg it recognised, FALSE if not
// Note: processing some arguments causes output to stdout to be generated.
//**********************************************************************

BOOL CNTService::ParseStandardArgs(int argc, char* argv[])
{ DWORD d;
  HKEY hkey;
  char s[256];

// See if we have any command line args we recognise

  if (argc <= 1)
    return FALSE;

// -v means print version info and status

  if (lstrcmp(argv[1], "-v") == 0)
  { wsprintf(s, "%s Version %d.%d\n", m_szServiceName, m_iMajorVersion, m_iMinorVersion);
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s, lstrlen(s), &d, NULL);
    wsprintf(s, "The service is %s installed\n", IsInstalled() ? "currently" : "not");
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s, lstrlen(s), &d, NULL);
    return TRUE;
  }

// -i means install

  else if (lstrcmp(argv[1], "-i") == 0)
  { if (IsInstalled())
    { wsprintf(s, "%s is already installed\n", m_szServiceName);
    }
    else
    { if (Install())
      { wsprintf(s, "%s installed\n", m_szServiceName);
      }
      else
      { wsprintf(s, "%s failed to install. Error %d\n", m_szServiceName, GetLastError());
      }
    }
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s, lstrlen(s), &d, NULL);
    return TRUE;
  }

// -u means uninstall

  else if (lstrcmp(argv[1], "-u") == 0)
  { if (!IsInstalled())
    { wsprintf(s, "%s is not installed\n", m_szServiceName);
    }
    else
    { if (Uninstall())
      { char szFilePath[_MAX_PATH];

        ::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));
        wsprintf(s, "%s removed. (You must delete the file (%s) yourself.)\n", m_szServiceName, szFilePath);
      }
      else
      { wsprintf(s, "Could not remove %s. Error %d\n", m_szServiceName, GetLastError());
      }
    }
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s, lstrlen(s), &d, NULL);
    return TRUE; // say we processed the argument
  }

// Anything else means set this to be the app path

  else
  { if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, PATHKEY, 0, "", 0, KEY_ALL_ACCESS, NULL, &hkey, &d) != ERROR_SUCCESS)
    { printf("Cannot open key %s\n%s\n", PATHKEY, GetLastErrorMessage());
      return TRUE;
    }

    if (RegSetValueEx(hkey, PATHVALUE, 0, REG_SZ, (BYTE*) argv[1], lstrlen(argv[1])) != ERROR_SUCCESS)
    { printf("Cannot set value %s/%s\n%s\n", PATHVALUE, argv[1], GetLastErrorMessage());
      return TRUE;
    }

    RegCloseKey(hkey);
    return TRUE;
  }

// Don't recognise the args

  return FALSE;
}


//**********************************************************************
// IsInstalled
// ===========
// Check if the service is installed
//**********************************************************************

BOOL CNTService::IsInstalled()
{ BOOL bResult = FALSE;

// Open the Service Control Manager

  SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

// Try to open the service to test if it is installed

  if (hSCM)
  { SC_HANDLE hService = ::OpenService(hSCM, m_szServiceName, SERVICE_QUERY_CONFIG);
    if (hService)
    { bResult = TRUE;
      ::CloseServiceHandle(hService);
    }

    ::CloseServiceHandle(hSCM);
  }

  return bResult;
}


//**********************************************************************
// Install
// =======
// Install ourself
//**********************************************************************

BOOL CNTService::Install()
{

// Open the Service Control Manager

  SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  if (!hSCM)
    return FALSE;

// Get the executable file path

  char szFilePath[MAX_PATH+1];
  ::GetModuleFileName(NULL, szFilePath, MAX_PATH);

// Create the service

  SC_HANDLE hService = ::CreateService(hSCM,
                                       m_szServiceName,
                                       m_szServiceName,
                                       SERVICE_ALL_ACCESS,
                                       SERVICE_WIN32_OWN_PROCESS,
                                       SERVICE_DEMAND_START, // start condition
                                       SERVICE_ERROR_NORMAL,
                                       szFilePath,
                                       NULL,
                                       NULL,
                                       NULL,
                                       NULL,
                                       NULL);

  if (!hService)
  { ::CloseServiceHandle(hSCM);
    return FALSE;
  }

// make registry entries to support logging messages
// Add the source name as a subkey under the Application
// key in the EventLog service portion of the registry.

  char szKey[256];
  HKEY hKey = NULL;

  lstrcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\");
  lstrcat(szKey, m_szServiceName);

  if (::RegCreateKey(HKEY_LOCAL_MACHINE, szKey, &hKey) != ERROR_SUCCESS)
  { ::CloseServiceHandle(hService);
    ::CloseServiceHandle(hSCM);
    return FALSE;
  }

// Add the Event ID message-file name to the 'EventMessageFile' subkey.

  ::RegSetValueEx(hKey, "EventMessageFile", 0, REG_EXPAND_SZ, (CONST BYTE*)szFilePath, lstrlen(szFilePath) + 1);

// Set the supported types flags.

  DWORD dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;
  ::RegSetValueEx(hKey, "TypesSupported", 0, REG_DWORD, (CONST BYTE*)&dwData, sizeof(DWORD));
  ::RegCloseKey(hKey);

  LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_INSTALLED, m_szServiceName);

// tidy up

  ::CloseServiceHandle(hService);
  ::CloseServiceHandle(hSCM);

  return TRUE;
}


//**********************************************************************
// Uninstall
// =========
// Uninstall ourself.  This does not delete the service file.
//**********************************************************************

BOOL CNTService::Uninstall()
{ BOOL bResult = FALSE;

// Open the Service Control Manager

  SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  if (!hSCM)
    return FALSE;

// Open the service

  SC_HANDLE hService = ::OpenService(hSCM, m_szServiceName, DELETE);

// Delete the service

  if (hService)
  { if (::DeleteService(hService))
    { LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_REMOVED, m_szServiceName);
      bResult = TRUE;
    }
    else
    { LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_NOTREMOVED, m_szServiceName);
    }
    ::CloseServiceHandle(hService);
  }

  ::CloseServiceHandle(hSCM);

  return bResult;
}


//**********************************************************************
// LogEvent
// ========
// This function makes an entry into the application event log
//**********************************************************************

void CNTService::LogEvent(WORD wType, DWORD dwID, const char* pszS1, const char* pszS2, const char* pszS3)
{ int iStr = 0;
  const char* ps[3];

  ps[0] = pszS1;
  ps[1] = pszS2;
  ps[2] = pszS3;

  for (int i = 0; i < 3; i++)
    if (ps[i] != NULL)
      iStr++;

// Check the event source has been registered and if not then register it now

  if (!m_hEventSource)
    m_hEventSource = ::RegisterEventSource(NULL, m_szServiceName);

  if (m_hEventSource)
  { ::ReportEvent(m_hEventSource,
                  wType,
                  0,
                  dwID,
                  NULL, // sid
                  iStr,
                  0,
                  ps,
                  NULL);
  }
}


//**********************************************************************
// StartService
// ============
// Start the service
//**********************************************************************

BOOL CNTService::StartService()
{
  SERVICE_TABLE_ENTRY st[] =
  {
    {m_szServiceName, ServiceMain},
    {NULL, NULL}
  };

  BOOL b = ::StartServiceCtrlDispatcher(st);

  return b;
}


//**********************************************************************
// ServiceMain
// ===========
// This is the function which is called to get the service going.
// After some initialisation it calls the service objects's Run method
// and this method actually does the work of the service.
//
// This is a static function which manipulates the static variable
// m_pThis which contains the one and only instance of the service.
//**********************************************************************

void CNTService::ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv)
{

// Get a pointer to the one and only service object

  CNTService* pService = m_pThis;

// Register the control request handler

  pService->m_Status.dwCurrentState = SERVICE_START_PENDING;
  pService->m_hServiceStatus = RegisterServiceCtrlHandler(pService->m_szServiceName, Handler);

  if (pService->m_hServiceStatus == NULL)
  { pService->LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_CTRLHANDLERNOTINSTALLED);
    return;
  }

// Start the initialisation.  When the Run function returns, the
// service has stopped.

  if (pService->Initialize())
  { pService->m_bIsRunning = TRUE;
    pService->m_Status.dwWin32ExitCode = 0;
    pService->m_Status.dwCheckPoint = 0;
    pService->m_Status.dwWaitHint = 0;
    pService->Run();
  }

// Tell the service manager we are stopped

  pService->SetStatus(SERVICE_STOPPED);
}


//**********************************************************************
// SetStatus
// =========
// Tell the service manager our status
//**********************************************************************

void CNTService::SetStatus(DWORD dwState)
{
  m_Status.dwCurrentState = dwState;
  ::SetServiceStatus(m_hServiceStatus, &m_Status);
}


//**********************************************************************
// Initialize
// ==========
// Initialise the object prior to starting
//**********************************************************************

BOOL CNTService::Initialize()
{

// Start the initialization

  SetStatus(SERVICE_START_PENDING);

// Perform the actual initialization.  The OnInit method is normally
// overriden by subclasses to do subclass specific initialisation.

  BOOL bResult = OnInit();

// Set final state

  m_Status.dwWin32ExitCode = GetLastError();
  m_Status.dwCheckPoint = 0;
  m_Status.dwWaitHint = 0;

// If initialisation failed stop now

  if (!bResult)
  { LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_FAILEDINIT);
    SetStatus(SERVICE_STOPPED);
    return FALSE;
  }

// Tell the service manager we have started

  LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STARTED);
  SetStatus(SERVICE_RUNNING);

  return TRUE;
}


//**********************************************************************
// Handler
// =======
// Static member function (callback) to handle commands from the
// service control manager
//
// This is a static function which manipulates the static variable
// m_pThis which contains the one and only instance of the service.
//**********************************************************************

void CNTService::Handler(DWORD dwOpcode)
{

// Get a pointer to the object

  CNTService* pService = m_pThis;

  switch (dwOpcode)
  {

// Stop the service

    case SERVICE_CONTROL_STOP:
      pService->SetStatus(SERVICE_STOP_PENDING);
      pService->OnStop();
      pService->m_bIsRunning = FALSE;
      pService->LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STOPPED);
      break;

// Pause the service

    case SERVICE_CONTROL_PAUSE:
      pService->OnPause();
      break;

// Continue the service

    case SERVICE_CONTROL_CONTINUE:
      pService->OnContinue();
      break;

// Ask the service for its current status

    case SERVICE_CONTROL_INTERROGATE:
      pService->OnInterrogate();
      break;

// The system is shutting down

    case SERVICE_CONTROL_SHUTDOWN:
      pService->OnShutdown();
      break;

// Here handle user defined and unrecognised control codes

    default:
      if (dwOpcode >= SERVICE_CONTROL_USER)
      { if (!pService->OnUserControl(dwOpcode))
        { pService->LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_BADREQUEST);
        }
      }
      else
      { pService->LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_BADREQUEST);
      }
      break;
  }

// Report current status

  ::SetServiceStatus(pService->m_hServiceStatus, &pService->m_Status);
}


//**********************************************************************
// Run
// ===
// This is where the service does its work.  Subclasses will normally
// override this function, the implementation here does nothing.
//**********************************************************************

void CNTService::Run()
{

  while (m_bIsRunning)
    Sleep(5000);
}


//**********************************************************************
// Virtual functions
// These are overridden by subclasses to take whatever action the
// subclass requires.
//**********************************************************************

// Called when the service is first initialized

BOOL CNTService::OnInit()
{
  return TRUE;
}

// Called when the service control manager wants to stop the service

void CNTService::OnStop()
{
}

// called when the service is interrogated

void CNTService::OnInterrogate()
{
}

// called when the service is paused

void CNTService::OnPause()
{
}

// called when the service is continued

void CNTService::OnContinue()
{
}

// called when the service is shut down

void CNTService::OnShutdown()
{
}

// called when the service gets a user control message

BOOL CNTService::OnUserControl(DWORD dwOpcode)
{
  return FALSE; // say not handled
}


//**********************************************************************
// GetLastErrorMessage
// -------------------
//**********************************************************************

const char* GetLastErrorMessage(void)
{ DWORD d;
  static char errmsg[1024];

  d = GetLastError();

  strcpy(errmsg, "<unknown error>");
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, d, 0, errmsg, 1023, NULL);
  return(errmsg);
}


