//**********************************************************************
// CRhsAdminSvc
// ============
//
// This class is a subclass of the CNTService class and implements the
// RhsAdminSvc server.
//
//
// John Rennie
// jrennie@cix.compulink.co.uk
// 5th July 1998
//**********************************************************************

#ifndef STRICT
#define STRICT
#endif

#include <windows.h>
#include <stdarg.h>
#include <Misc/RhsSecurityDescriptor.h>
#include "RhsAdminSvc.h"
#include "CRhsAdminSvc.h"

#include "GetLastErrorMessage.cpp"


//**********************************************************************
// Prototypes
// ----------
//**********************************************************************

DWORD WINAPI RunThread(LPVOID hPipe);

void logprintf(const char *Format, ...);


//**********************************************************************
// Global variables
// ----------------
//**********************************************************************

#define PATHKEY   "SYSTEM\\CurrentControlSet\\Services\\RhsAdminSvc\\Parameters"
#define PATHVALUE "Path"

#define IDS_String_NOPATH "Cannot run %s.\nThe RhsAdminSvc service is not properly installed: the Path parameter is missing."

#define IDS_String_RUN     "%s started."
#define IDS_String_CANTRUN "Cannot run %s.\n%s"


//**********************************************************************
// Constructor
// -----------
// Call the base class constructor passing our name
//**********************************************************************

CRhsAdminSvc::CRhsAdminSvc():CNTService("RhsAdminSvc")
{
  m_Listen = INVALID_HANDLE_VALUE;
  m_SD = NULL;
}


//**********************************************************************
// OnInit
// ------
// Called when the service is loaded.  We need to create all resources
// used by the service.
//**********************************************************************

BOOL CRhsAdminSvc::OnInit()
{ DWORD d;
  SECURITY_ATTRIBUTES sa;
  char s[128];

// Create the pipe used for accepting requests.  Use a security descriptor
// which allows everyone access.

  m_SD = new CRhsSecurityDescriptor;
  m_SD->InitFromCurrentProcess();
  m_SD->GrantRightsTo("Everyone", NULL, FILE_ALL_ACCESS, TRUE);

  sa.nLength = sizeof(sa);
  sa.bInheritHandle = FALSE;
  sa.lpSecurityDescriptor = m_SD->GetSD();

  m_Listen = CreateNamedPipe(RHSADMINSVCPIPE,
                             PIPE_ACCESS_DUPLEX,
                             PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE | PIPE_WAIT,
                             1,
                             0x1000, 0x1000,
                             10000,
                             &sa);

  if (m_Listen == INVALID_HANDLE_VALUE)
  { logprintf("CRhsAdminSvc::OnInit: Error creating pipe: %s", GetLastErrorMessage());
    return FALSE;
  }

// Log the startup

  d = 127;
  GetComputerName(s, &d);
  logprintf("Started on %s\r\n", s);

// Return indicating success

  return TRUE;
}


//**********************************************************************
// OnStop
// ------
// Remove the resources we created in OnInit
//**********************************************************************

void CRhsAdminSvc::OnStop()
{ DWORD d;
  HANDLE h;
  char s[128];

// Set the flag to show we must stop

  m_bIsRunning = FALSE;

// Now send ourselves an exit command

  h = CreateFile(RHSADMINSVCPIPE, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (h != INVALID_HANDLE_VALUE)
  { WriteFile(h, "exit", 4, &d, NULL);
    CloseHandle(h);
  }

// Log the shutdown

  d = 127;
  GetComputerName(s, &d);
  logprintf("Stopped on %s\r\n", s);
}


//**********************************************************************
// Run
// ---
// This is where the work is done.  We sit in a loop accepting commands
// from clients via the pipe.  The command is parsed and executed.
//**********************************************************************

void CRhsAdminSvc::Run(void)
{ DWORD d;
  char command[1024], username[32];

  while (m_bIsRunning)
  {

// Listen for client requests

    if (!ConnectNamedPipe(m_Listen, NULL))
    { logprintf("CRhsAdminSvc::Run: Error connecting pipe: %s", GetLastErrorMessage());
      Sleep(1000); // Avoid pathological looping
      continue;
    }

// Read the command

    if (!ReadFile(m_Listen, command, 1023, &d, NULL))
    { logprintf("CRhsAdminSvc::Run: Error reading from pipe: %s", GetLastErrorMessage());
      DisconnectNamedPipe(m_Listen);
      continue;
    }
    command[d] = '\0';

    GetNamedPipeHandleState(m_Listen, NULL, NULL, NULL, NULL, username, 31);
    logprintf("Accepted connection from %s: %s\r\n", username, command);

// Parse the command

    for (d = 0; command[d] != ' ' && command[d] != '\0'; d++);
    if (command[d] == ' ')
    { command[d++] = '\0';
      while (command[d] == ' ')
        d++;
    }

// exit: Exit the server

    if (lstrcmpi(command, "exit") == 0)
    { break;
    }

// exec: Run an application

    else if (lstrcmpi(command, "exec") == 0)
    { ExecApp(command + d);
    }

// Unrecognised command

    else
    { logprintf("CRhsAdminSvc::Run: Unrecognised command \"%s\"\r\n", command);
    }

// Disconnect this client and go back for the next

    DisconnectNamedPipe(m_Listen);
  }

// Close the pipe

  if (m_Listen != INVALID_HANDLE_VALUE)
  { CloseHandle(m_Listen);
    m_Listen = NULL;
  }

  if (m_SD)
  { delete m_SD;
    m_SD = NULL;
  }
}


//**********************************************************************
// CRhsAdminSvc::ExecApp
// ---------------------
// Run an application in the security context of the service and
// interactively.
//
// The client can supply only the executable name.  The path is read
// from the registry key:
//
// SYSTEM\CurrentControlSet\Services\RhsAdminSvc\Parameters
//
// This ensures that clients cannot run arbitrary applications.
//**********************************************************************

BOOL CRhsAdminSvc::ExecApp(const char* AppName)
{ DWORD type, d;
  HKEY hkey;
  STARTUPINFO sif;
  PROCESS_INFORMATION pif;
  char app[1024], s[1024];

// Prepend the app with the only path we are prepared to run apps from

  lstrcpy(app, "");

  RegCreateKeyEx(HKEY_LOCAL_MACHINE, PATHKEY, 0, "", 0, KEY_READ, NULL, &hkey, &d);
  d = 255;
  RegQueryValueEx(hkey, PATHVALUE, NULL, &type, (BYTE*) app, &d);
  RegCloseKey(hkey);

  if (lstrcmp(app, "") == 0)
  { logprintf("exec: Path not set in registry\r\n");
    wsprintf(s, IDS_String_NOPATH, AppName);
    WriteFile(m_Listen, s, lstrlen(s), &d, NULL);
    return FALSE;
  }

  lstrcat(app, "\\");
  lstrcat(app, AppName);

// Run the requested application

  sif.cb          = sizeof(STARTUPINFO);
  sif.lpReserved  = NULL;
  sif.lpDesktop   = NULL;
  sif.lpTitle     = NULL;
  sif.dwFlags     = STARTF_USESHOWWINDOW;
  sif.wShowWindow = SW_SHOWNORMAL;
  sif.cbReserved2 = 0;
  sif.lpReserved2 = NULL;

  if (!CreateProcess(NULL, app, NULL, NULL, FALSE, NORMAL_PRIORITY_CLASS, NULL, NULL, &sif, &pif))
  { d = GetLastError();
    logprintf("exec: Error starting %s: %s", app, GetLastErrorMessage(d));

    wsprintf(s, IDS_String_CANTRUN, AppName, GetLastErrorMessage(d));
    WriteFile(m_Listen, s, lstrlen(s), &d, NULL);

    return FALSE;
  }

  CloseHandle(pif.hThread);
  CloseHandle(pif.hProcess);

  logprintf("exec: %s executed\r\n", app);

  wsprintf(s, IDS_String_RUN, AppName);
  WriteFile(m_Listen, s, lstrlen(s), &d, NULL);

// Return indicating success

  return TRUE;
}


//**********************************************************************
// logprintf
// =========
// Log a message
//**********************************************************************

void logprintf(const char *Format, ...)
{ DWORD d;
  HANDLE h;
  SYSTEMTIME st;
  va_list ap;
  char s[1024];

  GetTempPath(1023, s);
  if (s[lstrlen(s)-1] != '\\')
    lstrcat(s, "\\");
  lstrcat(s, "RhsAdminSvc.log");

  h = CreateFile(s, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, 0, NULL);

  if (h != INVALID_HANDLE_VALUE)
  { SetFilePointer(h, 0, NULL, FILE_END);

    GetLocalTime(&st);
    wsprintf(s, "%02i/%02i/%04i %02i:%02i:%02i : ", st.wDay, st.wMonth, st.wYear, st.wHour, st.wMinute, st.wSecond);

    va_start(ap, Format);
    wvsprintf(s + lstrlen(s), Format, ap);
    va_end(ap);

    WriteFile(h, s, lstrlen(s), &d, NULL);
    CloseHandle(h);
  }
}


